from .base import Metaclass
from .storage import RedisTempStore
from django.db.models.signals import post_save, pre_delete

default_signals = {
    'post_save': post_save,
    'pre_delete': pre_delete
}


class Observer(object):

    __metaclass__ = Metaclass

    def get_app_label(self):
        return self._meta.model._meta.app_label

    def get_model_name(self):
        return self.get_model()

    def get_model(self):
        return self._meta.model

    def get_fields(self):
        """
        Get all watched fields (declared in Meta class)
        """
        return self._meta.fields


class ModelObserver(Observer):
    """
    Observer that listens to some built-in and custom signals.
    Receivers should be named as RECEIVERNAME_receiver
    where RECEIVERNAME is the name of the receiver.
        E.g: a receiver for 'post_save signal' would be 'post_save_receiver'
    """
    def __init__(self, *args, **kwargs):
        self.custom_signals = kwargs.get('custom_signals', {})

        self.custom_signals.update(default_signals)
        for signal_name, signal in self.custom_signals.items():
            signal_receiver = getattr(self, '%s_receiver' % signal_name, None)
            if signal_receiver:
                # weak=False stops the receiver from being garbage collected
                signal.connect(signal_receiver, sender=self.get_model(), weak=False)

    @property
    def signals(self):
        self.custom_signals.update(default_signals)
        return self.custom_signals

    def __delattr__(self, attr):
        for signal_name, signal in self.signals.items():
            signal_receiver = getattr(self, '%s_receiver' % signal_name, None)
            if signal_receiver:
                signal.disconnect(signal_receiver, sender=self.get_model())

    def post_save_receiver(self, sender, instance, created, **kwargs):
        self.publisher.on_instance_change(instance, created, **kwargs)


class ObjectObserver(ModelObserver):
    """
    Model instance observer that tracks changes of defined fields.
    """
    def __init__(self, storage=RedisTempStore, *args, **kwargs):
        super(ObjectObserver, self).__init__(*args, **kwargs)

        timeout = kwargs.get('tempstore_timeout', None)
        tempstore_settings = {
            'observer': self,
            'prefix': 'sherlock:tempstore:%s:%s' % (self.get_app_label(), self.get_model_name()),
        }
        if timeout:
            tempstore_settings.update({'timeout': timeout})

        self.storage_client = storage(**tempstore_settings)

    def get_instance_field_value(self, instance, field_name):
        """
        Get field value of an instance, handles ForeignKey using pk
        """
        field = self.get_model()._meta.get_field_by_name(field_name)[0]
        field_type = field.get_internal_type()
        field_value = getattr(instance, field_name, None)
        if field_type in ('ForeignKey', ):
            return getattr(instance, field_name).pk if field_value else field_value
        else:
            return getattr(instance, field_name) if field_value else field_value

    def post_save_receiver(self, sender, instance, created, **kwargs):
        if not created:
            for observed_field in self.get_fields():
                changes = self.storage_client.get_changes(instance, observed_field)
                if changes:
                    self.publisher.on_field_change(
                        instance, observed_field, changes=changes
                    )



from django.test import TestCase
from polls.models import Poll
from sherlock.models import Channel


class ChannelManagerTest(TestCase):

    def setUp(self):
        self.poll = Poll.objects.create(question='What is going on?')

    def tearDown(self):
        self.poll.delete()

    def test_construct_identifier_for_instance(self):
        identifier = Channel.objects.construct_identifier(instance=self.poll)
        self.assertEqual(identifier, 'app:polls:model:poll:instance:1')

    def test_construct_identifier_for_field(self):
        identifier = Channel.objects.construct_identifier(instance=self.poll, field='question')
        self.assertEqual(identifier, 'app:polls:model:poll:instance:1:field:question')

    def test_construct_identifier_for_non_existing_field(self):
        with self.assertRaises(AttributeError):
            Channel.objects.construct_identifier(instance=self.poll, field='foo')

    def test_create_field_for_instance(self):
        identifier = Channel.objects.create_channel(self.poll)
        self.assertEqual(identifier, 'app:polls:model:poll:instance:1')
        self.assertIsInstance(Channel.objects.get(name=identifier), Channel)

    def test_create_field_for_field(self):
        identifier = Channel.objects.create_channel(self.poll, 'question')
        self.assertEqual(identifier, 'app:polls:model:poll:instance:1:field:question')
        self.assertIsInstance(Channel.objects.get(name=identifier), Channel)



from .base import Metaclass
from .models import Channel, Subscriber


class BasePublisher(object):

    __metaclass__ = Metaclass

    def on_instance_change(self, instance, created, **kwargs):
        """
        Called when model observer receives a 'post_save' signal
        """
        identifier = self._create_channel(instance)

        publish_method = getattr(self, 'publish_instance_change', None)
        if publish_method:
            publish_method(self, instance, created)
        else:
            self._publish(instance, identifier, created=created)

    def on_field_change(self, instance, field, **kwargs):
        """
        Called when an observed field has changed
        """
        changes = kwargs.get('changes')
        publish_method = getattr(self, 'publish_%s' % field, self._publish)
        identifier = self._create_channel(instance, field)
        if not self._requires_authorisation(field):
            publish_method(instance, identifier, field=field, changes=changes)
        else:
            self.authorise(instance, *changes)

    def _create_channel(self, instance, field=None):
        return Channel.objects.create_channel(instance, field)

    def _requires_authorisation(self, field):
        return hasattr(self._meta, 'requires_authorisation') \
            and field in self._meta.requires_authorisation

    def authorise(self, instance, *changes):
        # TODO: figure out the authorisation process
        print '==============> Authorising instance pk: %s' % instance.pk

    def _publish(self, instance, identifier, **context):
        """
        Figures out details of subscribers and publishes the notification.
        send_email() method has to be implemented
        """
        if self._meta.email:
            subscribers = Subscriber.objects.filter(channels__name__in=[identifier])
            emails = [sub.email for sub in subscribers]
            self.send_email(emails, instance, **context)

    def send_email(subscribers, instance, field=None, changes=None):
        """
        Sends emails to all subscribers, context is provided
        """
        raise NotImplementedError



from sherlock.publishers import BasePublisher


class PollPublisher(BasePublisher):
    class Meta:
        # requires_authorisation = ('question', )
        email = True

    def send_email(self, emails, instance, **context):
        print '==============>', self
        print '==============> Sending email to: %s' % emails
        print '==============> Instance pk: %s' % instance.pk
        print '==============> Context: %s' % context

    def publish_question(self, instance, identifier, **context):
        print '==============>', self
        print '==============> Instance pk: %s' % instance.pk
        print '==============> Identifier: %s' % identifier
        print '==============> Context: %s' % context



# from sherlock.observers import ObjectObserver, ModelObserver
# from .models import Poll, Choice
# from .signals import test_signal
# from .publishers import PollPublisher

# custom_signals = {
#     'test_signal': test_signal
# }


# class PollModelObserver(ModelObserver):
#     publisher = PollPublisher()

#     class Meta:
#         model = Poll

#     def test_signal_receiver(self, sender, **kwargs):
#         print '==============>', self
#         print '==============> Sender: %s' % sender
#         print '==============> Objects: %s' % kwargs['objects']


# class PollObserver(ObjectObserver):
#     publisher = PollPublisher()

#     class Meta:
#         model = Poll
#         fields = ('question', )


# class ChoiceObserver(ObjectObserver):

#     class Meta:
#         model = Choice
#         fields = ('poll', )

# poll_model_observer = PollModelObserver(custom_signals=custom_signals)
# poll_observer = PollObserver()
# choice_observer = ChoiceObserver()





class method_logger(object):
    def __init__(self, method):
        self.method = method
        self.was_called = False
        self.context = {}

    def __call__(self, *args, **kwargs):
        self.method(*args, **kwargs)
        self.was_called = True
        self.context = {
            'args': args,
            'kwargs': kwargs
        }





class Option(object):

    def __new__(cls, meta=None):
        # Return a new class base on ourselves
        attrs = dict(
            (name, getattr(meta, name))
            for name in dir(meta)
            if not name[0] == '_'
        )
        return object.__new__(type('Option', (cls,), attrs))


class Metaclass(type):

    def __new__(cls, name, bases, attrs):
        observer_class = super(Metaclass, cls).__new__(cls, name, bases, attrs)
        opts = getattr(observer_class, 'Meta', None)
        observer_class._meta = Option(opts)

        return observer_class



import redis
from django.conf import settings
try:
    import cPickle as pickle
except ImportError:
    import pickle


class BaseStorage(object):

    def __init__(self, *arg, **kwargs):
        self.observer = kwargs.get('observer', None)

    def get(self, instance, field_name):
        """
        Gets field value from temporary storage.
        """
        raise NotImplementedError

    def set(self, instance, field_name, value):
        """
        Temporarily stores field value. Timout can be define.
        Used in comparison when the the field changes.
        """
        raise NotImplementedError

    def get_changes(self, observer, instance, field_name):
        """
        Compare previous and current value of the field.
        Return previous and current value in a list if there are changes:
        dict(
            previous='previous_value',
            current='current_value'
        )
        """
        raise NotImplementedError


class HashTempStore(object):
    """
    key/field                    | instance1.pk | instance2.pk |
    prefix:app:model:field1      |    value1    |    value2    |
    prefix:app:model:field2      |    value1    |    value2    |
    """

    def __init__(self, *args, **kwargs):
        self.timeout = kwargs.get('timeout', 60 * 60 * 24)
        self.prefix = kwargs.get('prefix', None)
        self._client = redis.StrictRedis(**settings.REDIS_SETTINGS)
        super(HashTempStore, self).__init__(*args, **kwargs)

    def make_hash_key(self, key):
        return '%s:%s' % (self.prefix, key)

    def get(self, key, field):
        hash_key = self.make_hash_key(key)
        value = self._client.hget(hash_key, field)
        if value is None:
            return
        return pickle.loads(value)

    def set(self, key, mapping):
        assert isinstance(mapping, dict)
        hash_key = self.make_hash_key(key)
        mapping = dict([(m[0], pickle.dumps(m[1])) for m in mapping.items()])
        self._client.hmset(hash_key, mapping)
        self._client.expire(hash_key, self.timeout)


class RedisTempStore(HashTempStore, BaseStorage):
    """
    Wrapper class around HashTempStore to interface with observer class
    """

    def get(self, instance, field_name):
        return super(RedisTempStore, self).get(field_name, str(instance.pk))

    def set(self, instance, field_name, value):
        hash_field_name = str(instance.pk)
        super(RedisTempStore, self).set(
            field_name, {hash_field_name: value}
        )

    def get_changes(self, instance, field_name):
        previous = self.get(instance, field_name)
        current = self.observer.get_instance_field_value(instance, field_name)
        if previous != current:
            self.set(instance, field_name, current)
            return {
                'previous': previous,
                'current': current
            }
        return None




from django.contrib import admin
from polls import models
from .signals import test_signal


def send_signal(modeladmin, request, queryset):
    test_signal.send_robust(
        sender=modeladmin.model,
        objects=queryset
    )


class PollAdmin(admin.ModelAdmin):
    list_display = ('question', 'pub_date')

    actions = (send_signal, )

admin.site.register(models.Poll, PollAdmin)
admin.site.register(models.Choice)



from django.test import TestCase
from polls.models import Poll
from sherlock.observers import ModelObserver, ObjectObserver
from sherlock.publishers import BasePublisher
from sherlock.models import Channel
from .method_logger import method_logger


class ModelObserverTest(TestCase):

    def setUp(self):

        class PollPublisher(BasePublisher):
            class Meta:
                email = True

            def send_email(self, emails, instance, **context):
                pass

        class PollModelObserver(ModelObserver):
            publisher = PollPublisher()

            class Meta:
                model = Poll

        self.observer = PollModelObserver()
        self.publisher = self.observer.publisher
        self.publisher.send_email = method_logger(self.publisher.send_email)
        self.publisher.on_instance_change = method_logger(self.publisher.on_instance_change)

    def tearDown(self):
        del self.observer
        del self.publisher


    def test_channel_created(self):
        poll = Poll.objects.create(question='What is going on?')
        identifier = Channel.objects.construct_identifier(instance=poll)
        self.assertIsInstance(Channel.objects.get(name=identifier), Channel)
        poll.delete()

    def test_send_email(self):
        poll = Poll.objects.create(question='What is going on?')
        self.assertTrue(self.publisher.send_email.was_called)
        poll.delete()

    def test_send_email_not_called(self):
        self.publisher._meta.email = False
        poll = Poll.objects.create(question='What is going on?')
        self.assertFalse(self.publisher.send_email.was_called)
        poll.delete()

    def test_send_email_context(self):
        poll = Poll.objects.create(question='What is going on?')
        expected_context_for_create = {
            'args': ([], poll),
            'kwargs': {'created': True}
        }
        expected_context_for_update = {
            'args': ([], poll),
            'kwargs': {'created': False}
        }
        self.assertEqual(self.publisher.send_email.context, expected_context_for_create)
        poll.question = 'New question?'
        poll.save()
        self.assertEqual(self.publisher.send_email.context, expected_context_for_update)
        poll.delete()

    def test_on_instance_change(self):
        poll = Poll.objects.create(question='What is going on?')
        self.assertTrue(self.publisher.on_instance_change.was_called)
        poll.delete()




class ObjectObserverTest(TestCase):

    def test_publish_default(self):

        class PollPublisher(BasePublisher):
            class Meta:
                email = True

            def send_email(self, emails, instance, **context):
                pass

        class PollObserver(ObjectObserver):
            publisher = PollPublisher()

            class Meta:
                model = Poll
                fields = ('question', )

        self.observer = PollObserver()
        self.publisher = self.observer.publisher
        self.publisher.send_email = method_logger(self.publisher.send_email)
        self.publisher.on_field_change = method_logger(self.publisher.on_field_change)

        poll = Poll.objects.create(question='What is going on?')

        old_question = 'Old question?'
        new_question = 'New question?'

        poll.question = old_question
        poll.save()

        poll.question = new_question
        poll.save()

        self.assertTrue(self.publisher.on_field_change.was_called)
        self.assertEqual(self.publisher.send_email.context['args'], ([], poll))
        kwargs = {
            'field': 'question',
            'changes': {
                'current': new_question,
                'previous': old_question
            }
        }
        for k, v in kwargs.items():
            self.assertIn(k, self.publisher.send_email.context['kwargs'])
            self.assertEqual(v, self.publisher.send_email.context['kwargs'][k])

        poll.delete()
        del self.observer.signals


    def test_publish_custom(self):

        class PollPublisher(BasePublisher):

            def publish_question(self, instance, identifier, **context):
                pass

        class PollObserver(ObjectObserver):
            publisher = PollPublisher()

            class Meta:
                model = Poll
                fields = ('question', )

        self.observer = PollObserver()
        self.publisher = self.observer.publisher
        self.publisher.publish_question = method_logger(self.publisher.publish_question)
        self.publisher.on_field_change = method_logger(self.publisher.on_field_change)

        poll = Poll.objects.create(question='What is going on?')

        old_question = 'Old question?'
        new_question = 'New question?'

        poll.question = old_question
        poll.save()

        poll.question = new_question
        poll.save()

        self.assertTrue(self.publisher.on_field_change.was_called)
        self.assertEqual(
            self.publisher.publish_question.context['args'],
            (poll, Channel.objects.construct_identifier(instance=poll, field='question'))
        )
        kwargs = {
            'field': 'question',
            'changes': {
                'current': new_question,
                'previous': old_question
            }
        }
        for k, v in kwargs.items():
            self.assertIn(k, self.publisher.publish_question.context['kwargs'])
            self.assertEqual(v, self.publisher.publish_question.context['kwargs'][k])

        poll.delete()
        del self.observer.signals



from setuptools import setup, find_packages


VERSION = __import__("sherlock").__version__

CLASSIFIERS = [
    'Framework :: Django',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: BSD License',
    'Operating System :: OS Independent',
    'Topic :: Software Development',
]

install_requires = [
    'Django>=1.4.0',
    'redis'
]

setup(
    name="django-sherlock",
    description="A pluggable app for notifications",
    version=VERSION,
    author="Nam Ngo",
    author_email="nam@namis.me",
    url="https://github.com/geeknam/django-sherlock",
    packages=find_packages(exclude=["tests"]),
    install_requires=install_requires,
    classifiers=CLASSIFIERS,
)



from django.db import models


class ChannelManager(models.Manager):

    def construct_identifier(self, **kwargs):
        instance = kwargs.get('instance', None)
        field = kwargs.get('field', None)

        app_label = instance.__class__._meta.app_label
        model_name = instance.__class__._meta.object_name.lower()

        identifier = 'app:%s:model:%s:instance:%s'
        variables = [app_label, model_name, instance.pk]
        if field:
            if not hasattr(instance, field):
                raise AttributeError
            identifier += ':field:%s'
            variables.append(field)
        variables = tuple(variables)
        return identifier % variables

    def create_channel(self, instance, field=None):
        identifier = self.construct_identifier(instance=instance, field=field)
        self.get_or_create(name=identifier)

        return identifier



from django.conf import settings
from django.db import models
from .managers import ChannelManager


class Channel(models.Model):

    name = models.CharField(
        max_length=250,
        help_text='Identifier: app:polls:model:poll:instance:1:field:question',
        unique=True, db_index=True

    )

    objects = ChannelManager()

    def __unicode__(self):
        return self.name

    # TODO Find a better way to validate whether a channel is model, instance or field related
    def get_model(self):
        identifiers = self.name.split(':')
        return models.get_model(identifiers[1], identifiers[3])

    def get_instance(self):
        identifiers = self.name.split(':')
        if self.is_instance():
            return self.get_model().objects.get(pk=identifiers[5])

    def is_model(self):
        return len(self.name.split(':')) == 4

    def is_instance(self):
        return len(self.name.split(':')) == 6

    def is_field(self):
        return len(self.name.split(':')) == 8


class Subscriber(models.Model):

    email = models.EmailField(unique=True)
    channels = models.ManyToManyField('Channel')

    def __unicode__(self):
        return self.email


for app in settings.INSTALLED_APPS:
    try:
        module_name = '%s.observers' % str(app)
        __import__(module_name)
    except ImportError:
        pass




from django.db import models


class Poll(models.Model):
    question = models.CharField(max_length=200)
    pub_date = models.DateTimeField(auto_now_add=True)

    def __unicode__(self):
        return self.question


class Choice(models.Model):
    poll = models.ForeignKey(Poll)
    choice_text = models.CharField(max_length=200)
    votes = models.IntegerField()

    def __unicode__(self):
        return '%s - %s' % (self.poll.question, self.choice_text)
